unit WebModuleUnit1;

{===============================================================================
 Demo Web Service Part 3

 For Blaise Pascal Magazine 2021
 https://www.blaisepascalmagazine.eu/

 Author: Danny Wind (https://dannywind.nl)
 License: Creative Commons CC-BY

===============================================================================}

interface

uses
  System.SysUtils, System.Classes, System.Types, Web.HTTPApp;

type
  TWebModule1 = class(TWebModule)
    procedure WebModule1DefaultHandlerAction(Sender: TObject;
      Request: TWebRequest; Response: TWebResponse; var Handled: Boolean);
    procedure WebModule1WebActionItemNumberGetAction(Sender: TObject;
      Request: TWebRequest; Response: TWebResponse; var Handled: Boolean);
    procedure WebModule1WebActionItemKeyValueGETAction(Sender: TObject;
      Request: TWebRequest; Response: TWebResponse; var Handled: Boolean);
    procedure WebModule1WebActionItemKeyValuePUTAction(Sender: TObject;
      Request: TWebRequest; Response: TWebResponse; var Handled: Boolean);
    procedure WebModule1WebActionItemKeyValuePOSTAction(Sender: TObject;
      Request: TWebRequest; Response: TWebResponse; var Handled: Boolean);
    procedure WebModule1WebActionItemKeyValueDELETEAction(Sender: TObject;
      Request: TWebRequest; Response: TWebResponse; var Handled: Boolean);
    procedure WebModule1WebActionItemJavaScriptAction(Sender: TObject;
      Request: TWebRequest; Response: TWebResponse; var Handled: Boolean);
  private
    { Private declarations }
    function GetParameters(const aActionPath, aRequestPath: string): TStringDynArray;
  public
    { Public declarations }
  end;

var
  WebModuleClass: TComponentClass = TWebModule1;

implementation

{%CLASSGROUP 'Vcl.Controls.TControl'}

{$R *.dfm}

uses
  System.StrUtils, System.Generics.Collections, System.SyncObjs;

var
  gLock: TObject;
  gKeyValueStore: TDictionary<string, string>;

function TWebModule1.GetParameters(const aActionPath,
  aRequestPath: string): TStringDynArray;
var
  lActionPathLength: Integer;
  lRequestPathLength: Integer;
  lParameter: string;
  lParameters: TStringDynArray;
begin
  SetLength(Result, 0);
  lActionPathLength := aActionPath.Length;
  lRequestPathLength := aRequestPath.Length;
  if (lRequestPathLength > lActionPathLength) then
  begin
    lParameter := RightStr(aRequestPath, lRequestPathLength - lActionPathLength);
    lParameters := SplitString(lParameter,'/');
    if (Length(lParameters) > 0) then
    begin
      Result := lParameters;
    end
  end;
end;


procedure TWebModule1.WebModule1DefaultHandlerAction(Sender: TObject;
  Request: TWebRequest; Response: TWebResponse; var Handled: Boolean);
begin
  Response.Content :=
    '<html>' +
    '<head><title>Web Server Application</title></head>' +
    '<body>Web Server Application</body>' +
    '</html>';
end;

procedure TWebModule1.WebModule1WebActionItemJavaScriptAction(Sender: TObject;
  Request: TWebRequest; Response: TWebResponse; var Handled: Boolean);
begin
  Response.ContentType := 'text/html; charset=' + TEncoding.UTF8.MIMEName;
  Response.Content :=
    '<html>' +
    '<head><title>Call Number with JavaScript</title></head>' +
    '<body>Call Number with JavaScript ' +
    '<button onclick="getNumber()">Get Number in Console Log (view Ctrl-Shift-I).</button>' +
    '<script type="text/javascript">' +
    'function getNumber() {' +
    ' let url = ''http://localhost:8080/Number'';' +
    ' fetch(url).then(resp=> resp.json().then(j=> console.log(''\nNumber: '', j)));' +
    '}' +
    '</script>' +
    '</body>' +
    '</html>';
end;

procedure TWebModule1.WebModule1WebActionItemKeyValueDELETEAction(
  Sender: TObject; Request: TWebRequest; Response: TWebResponse;
  var Handled: Boolean);
{DELETE - "Delete"}
var
  lParameters: TStringDynArray;
  lKey: string;
begin
  lParameters := GetParameters((Sender as TWebActionItem).PathInfo, Request.PathInfo);
  if (Length(lParameters) > 0) then
  begin
    lKey := lParameters[0];
    Response.ContentType := 'application/json; charset=' + TEncoding.UTF8.MIMEName;
    if TMonitor.Enter(gLock, 500) then
    begin
      try
        if gKeyValueStore.ContainsKey(lKey) then
        begin
          gKeyValueStore.Remove(lKey);
          Response.Content := ' {"result":["OK"]}';
        end
        else
        begin
          Response.Content := '{"error":"Item not found"}';
        end;
      finally
        TMonitor.Exit(gLock);
      end;
    end;
    Handled := True;
  end
  else {No parameters on URL for GET request}
    Handled := False;
end;

procedure TWebModule1.WebModule1WebActionItemKeyValueGETAction(Sender: TObject;
  Request: TWebRequest; Response: TWebResponse; var Handled: Boolean);
{GET  - "Select" / Idempotent}
var
  lParameters: TStringDynArray;
  lKey: string;
  lValue: string;
begin
  lKey := '';
  {parse URL query parameters - http//localhost:8080/KeyValue?key=0}
  lKey := Request.QueryFields.Values['key'];
  if (lKey.IsEmpty) then
  begin {parse URL segment parameters - http//localhost:8080/KeyValue/0}
    lParameters := GetParameters((Sender as TWebActionItem).PathInfo,
                                 Request.PathInfo);
    if (Length(lParameters) > 0) then
    begin
      lKey := lParameters[0];
    end;
  end;
  if not(lKey.IsEmpty) then
  begin
    Response.ContentType := 'application/json; charset=' + TEncoding.UTF8.MIMEName;
    if TMonitor.Enter(gLock, 500) then
    begin
      try
        gKeyValueStore.TryGetValue(lKey, lValue);
      finally
        TMonitor.Exit(gLock);
      end;
    end;
    if not(lValue.IsEmpty) then
    begin
      // {"result":[JSONValue]}
      Response.Content := '{"result":[' + lValue + ']}'
    end
    else
    begin
      // {"error":"Item not found"}
      Response.Content := '{"error":"Item not found"}';
    end;
    Handled := True;
  end
  else {No parameters on URL for GET request}
    Handled := False;
end;

procedure TWebModule1.WebModule1WebActionItemKeyValuePOSTAction(Sender: TObject;
  Request: TWebRequest; Response: TWebResponse; var Handled: Boolean);
{POST - "Update"}
var
  lParameters: TStringDynArray;
  lKey: string;
  lValue: string;
begin
  lParameters := GetParameters((Sender as TWebActionItem).PathInfo, Request.PathInfo);
  if (Length(lParameters) > 0) then
  begin
    lKey := lParameters[0];
    lValue := '';
    if (Length(lParameters) > 1) then
    begin {Value is part or URL and a simple string}
      lValue := lParameters[1]
    end
    else
    begin {Value is send as content in the request and possibly a JSON or other complex string}
      lValue := Request.Content;
    end;
    if not(lValue.IsEmpty) then
    begin
      Response.ContentType := 'application/json; charset=' + TEncoding.UTF8.MIMEName;
      if TMonitor.Enter(gLock, 500) then
      begin
        try
          if gKeyValueStore.ContainsKey(lKey) then
          begin
            gKeyValueStore[lKey] := lValue;
            Response.Content := '{"result":["OK"]}';
          end
          else
          begin
            Response.Content := '{"error":"Item not found"}';
          end;
        finally
          TMonitor.Exit(gLock);
        end;
      end;
      Handled := True;
    end
    else
    begin
      Handled := False;
    end;
  end
  else
  begin
    {Do not reply}
    Handled := False;
  end;
end;


procedure TWebModule1.WebModule1WebActionItemKeyValuePUTAction(Sender: TObject;
  Request: TWebRequest; Response: TWebResponse; var Handled: Boolean);
{PUT  - "Insert or Update" / Idempotent}
var
  lParameters: TStringDynArray;
  lKey: string;
  lValue: string;
begin
  lParameters := GetParameters((Sender as TWebActionItem).PathInfo, Request.PathInfo);
  if (Length(lParameters) > 0) then
  begin
    lKey := lParameters[0];
    lValue := '';
    if (Length(lParameters) > 1) then
    begin {Value is part or URL and a simple string}
      lValue := lParameters[1]
    end
    else
    begin {Value is send as content in the request and possibly a JSON or other complex string}
      lValue := Request.Content;
    end;
    if not(lValue.IsEmpty) then
    begin
      Response.ContentType := 'application/json; charset=' + TEncoding.UTF8.MIMEName;
      if TMonitor.Enter(gLock, 500) then
      begin
        try
          gKeyValueStore.AddOrSetValue(lKey, lValue);
        finally
          TMonitor.Exit(gLock);
        end;
      end;
      Response.Content := '{"result":["OK"]}';
      Handled := True;
    end
    else
    begin
      Handled := False;
    end;
  end
  else
  begin
    {Do not reply}
    Handled := False;
  end;
end;

procedure TWebModule1.WebModule1WebActionItemNumberGetAction(Sender: TObject;
  Request: TWebRequest; Response: TWebResponse; var Handled: Boolean);
begin
  Response.ContentType := 'application/json; charset=' + TEncoding.UTF8.MIMEName;
  Response.Content := Random(100).ToString;
end;

initialization

  gLock := TObject.Create;
  gKeyValueStore := TDictionary<string, string>.Create;
  gKeyValueStore.AddOrSetValue('0', '"Zero"');

end.
